# Built-In Modules

# external Modules
from flask import Blueprint, render_template, request, session

from autonomous import log
from autonomous.auth import auth_required
from autonomous.auth.autoauth import AutoAuth

index_page = Blueprint("index", __name__)


@index_page.route("/", methods=("GET",))
def index():
    return render_template("index.html")


@index_page.route(
    "/protected",
    methods=(
        "GET",
        "POST",
    ),
)
@auth_required
def protected():
    if request.form:
        session.update(request.json)
    context = {"user": AutoAuth.current_user(), **request.json}
    return render_template("index.html", **context)



import sys

#from app.models.model import Model

from autonomous import log

class TestApp:
    def test_model(self):
        pass
        # m = Model(name="test")
        # m.save()
        # log(m.pk)
        # assert m.pk



# Built-In Modules
from datetime import datetime

# external Modules
from autonomous import log
from autonomous.tasks import AutoTasks
from flask import Blueprint, request

task_endpoint = Blueprint("task", __name__)


@task_endpoint.route("/checktask", methods=("POST",))
def checktask():
    # log(request.json.get("id"))
    task = AutoTasks().get_task(request.json.get("id"))
    log(task.result, task.status)
    return {"status": task.status, **task.result}



import time

from autonomous import log


def mocktask():
    log("MockTask - starting")
    time.sleep(1)
    log("MockTask - complete")
    return "success"


def longmocktask():
    log("LongTask - Starting")
    time.sleep(30)
    log("LongTask - Complete")
    return "success"


def parametermocktask(*args, **kwargs):
    log("ParameterMockTask - Starting", args, kwargs)
    result = args[0] + args[1]
    log("ParameterMockTask - complete", result)
    return result


def errormocktask(a, b):
    time.sleep(2)
    return int(a) + str(b)



import logging
import os

import rq_dashboard
from config import APIConfig
from flask import Flask, url_for
from views.api.task import task_endpoint


def create_api():
    app = Flask(os.getenv("APP_NAME", __name__))
    app.config.from_object(APIConfig)

    #################################################################
    #                             Plug-ins                          #
    #################################################################
    rq_dashboard.web.setup_rq_connection(app)

    ######################################
    #           Blueprints               #
    ######################################
    app.register_blueprint(rq_dashboard.blueprint, url_prefix="/rq")
    app.register_blueprint(task_endpoint, url_prefix="/task")

    ## Disables the favicon.ico not found error
    @app.route("/favicon.ico")
    def favicon():
        return url_for("static", filename="data:,")

    return app



# Built-In Modules

# external Modules
from flask import Blueprint, render_template, request, session
from autonomous import log
from autonomous.auth import auth_required

admin_page = Blueprint("admin", __name__)


@admin_page.route("/", methods=("GET",))
@auth_required
def index():
    if request.form:
        session.update(request.form)
    return render_template("admin.html")



# Built-In Modules

# external Modules
from datetime import datetime, timezone

from flask import Blueprint, redirect, render_template, request, session, url_for

from autonomous import log
from autonomous.auth import AutoUser, GithubAuth, GoogleAuth

auth_page = Blueprint("auth", __name__)


@auth_page.route("/login", methods=("GET", "POST"))
def login():
    # session["user"] = None
    # AutoUser.table().flush_table()
    if session.get("user"):
        if last_login := session["user"]["last_login"]:
            log(last_login)
            diff = datetime.now(timezone.utc) - last_login
            if diff.days <= 30 and session["user"]["state"] == "authenticated":
                return redirect(url_for("index.protected"))

    if request.method == "POST":
        session["user"] = None
        if request.form.get("authprovider") == "google":
            authorizer = GoogleAuth()
            session["authprovider"] = "google"
        elif request.form.get("authprovider") == "github":
            authorizer = GithubAuth()
            session["authprovider"] = "github"
        uri, state = authorizer.authenticate()
        session["authprovider_state"] = state
        # log(uri, state)
        return redirect(uri)

    return render_template("login.html")


@auth_page.route("/authorize", methods=("GET", "POST"))
def authorize():
    # log(request.args)
    if session["authprovider"] == "google":
        authorizer = GoogleAuth()
    elif session["authprovider"] == "github":
        authorizer = GithubAuth()
    response = str(request.url)
    # log(response)
    user_info, token = authorizer.handle_response(
        response, state=request.args.get("state")
    )
    # log(user_info["email"])
    user_info["provider"] = session["authprovider"]
    user = AutoUser.authenticate(user_info, token)
    # log(user)
    session["user"] = user.serialize()
    return redirect(url_for("auth.login"))


@auth_page.route("/logout", methods=("POST", "GET"))
def logout():
    if session.get("user"):
        user = AutoUser.get(session["user"]["pk"])
        user.state = "unauthenticated"
        user.save()
        session.pop("user")
        # log(f"User {user.name} logged out")
    return redirect(url_for("auth.login"))



from autonomous import log
from autonomous.model import AutoModel, AutoAttribute


class Model(AutoModel):
    # set model default attributes
    attributes = {
        "name": AutoAttribute("TEXT", default=""),
        "age": None,
    }



# opt : Optional[str]  # for optional attributes
# default : Optional[str] = "value"  # for default values
import importlib
import json
from abc import ABC
from datetime import datetime

from autonomous import log

from .autoattribute import AutoAttribute
from .orm import ORM


class DelayedModel:
    def __init__(self, model, pk):
        # log(model, pk)
        assert model
        assert pk
        module_name, class_name = model.rsplit(".", 1)
        module = importlib.import_module(module_name)
        model = getattr(module, class_name)
        object.__setattr__(self, "_delayed_model", model)
        object.__setattr__(self, "_delayed_pk", pk)
        object.__setattr__(self, "_delayed_obj", None)

    def _instance(self):
        if not object.__getattribute__(self, "_delayed_obj"):
            _pk = object.__getattribute__(self, "_delayed_pk")
            _model = object.__getattribute__(self, "_delayed_model")
            # log(_pk, _model)
            # breakpoint()
            _obj = _model(pk=_pk)
            # log(_obj)
            try:
                assert _obj
            except AssertionError as e:
                msg = f"{e}\n\nModel relationship error. Most likely failed to clean up dangling reference.\nModel: {_model}\npk: {_pk}\nResult: {_obj}"
                log(msg)
                raise Exception(msg)
            else:
                object.__setattr__(self, "_delayed_obj", _obj)

        return object.__getattribute__(self, "_delayed_obj")

    # def __getattr__(self, name):
    #     return getattr(self._instance(), name)

    def __getattribute__(self, name):
        # log(name)
        if name in [
            "_delayed_model",
            "_delayed_pk",
            "_delayed_obj",
            "_instance",
        ]:
            return object.__getattribute__(self, name)
        return object.__getattribute__(self._instance(), name)

    def __setattr__(self, name, value):
        if name.startswith("_delayed"):
            object.__setattr__(self, name, value)
        else:
            setattr(self._instance(), name, value)

    def __delattr__(self, name):
        delattr(self._instance(), name)

    def __nonzero__(self):
        return bool(self._instance())

    def __str__(self):
        return str(self._instance())

    def __repr__(self):
        result = repr(self._delayed_obj)
        msg = f"\n<<DelayedModel {self._delayed_model.__name__}:{self._delayed_pk}>>:\t{result}"
        return msg

    def __hash__(self):
        return hash(self._instance())


class AutoModel(ABC):
    attributes = {}
    _table_name = ""
    _table = None
    _orm = ORM

    __save_memo = []

    def __new__(cls, *args, **kwargs):
        """
        Create a new instance of the AutoModel.

        This method is responsible for creating a new instance of the AutoModel class.
        It sets default attributes, populates the object from the database if a primary key is provided,
        and handles additional keyword arguments.

        Args:
            cls: The class itself.
            *args: Positional arguments.
            **kwargs: Keyword arguments, including 'pk' for primary key.

        Returns:
            obj: The created AutoModel instance.
        """
        obj = super().__new__(cls)
        # set default attributes
        # Get model data from database
        result = cls.table().get(kwargs.get("pk")) or {}
        # set object attributes
        for k, v in cls.attributes.items():
            # set attribute from db or set to default value
            if isinstance(v, tuple):
                setattr(obj, k, result.get(k, v[0]))
            elif isinstance(v, AutoAttribute):
                setattr(obj, k, result.get(k, v.default))
            else:
                setattr(obj, k, result.get(k, v))

        # update model with keyword arguments
        obj.__dict__ |= kwargs

        # log(obj, kwargs)
        _automodel = obj.__dict__.pop("_automodel", None)
        obj.__dict__ |= cls._deserialize(obj.__dict__)
        obj.__dict__["_automodel"] = _automodel

        obj.last_updated = datetime.now()

        return obj

    @property
    def _save_memo(self):
        return AutoModel.__save_memo

    @_save_memo.setter
    def _save_memo(self, value):
        AutoModel.__save_memo = value

    def __repr__(self) -> str:
        """
        Return a string representation of the AutoModel instance.

        Returns:
            str: A string representation of the AutoModel instance.
        """
        return f"{self.__dict__}"

    def __getattribute__(self, name):
        obj = super().__getattribute__(name)
        if isinstance(obj, DelayedModel):
            return obj._instance()
        return obj

    @classmethod
    def table(cls):
        # breakpoint()
        if not cls._table or cls._table.name != cls.__name__:
            cls.attributes["pk"] = AutoAttribute("TAG", primary_key=True)
            cls.attributes["last_updated"] = datetime.now()
            cls.attributes["_automodel"] = AutoAttribute(
                "TAG", default=cls.model_name()
            )
            cls._table = cls._orm(cls._table_name or cls.__name__, cls.attributes)
        return cls._table

    @classmethod
    def model_name(cls):
        """
        Get the fully qualified name of this model.

        Returns:
            str: The fully qualified name of this model.
        """
        return f"{cls.__module__}.{cls.__name__}"

    ## TODO: Save all sub objects
    ## FIXME: This is causing a circular reference or no saving at all
    def _subobj_save(self, obj):
        if isinstance(obj, list):
            for v in obj:
                if self._subobj_save(v) is None:
                    obj.remove(v)
        elif isinstance(obj, dict):
            for k, v in obj.items():
                if self._subobj_save(v) is None:
                    obj[k] = None
        elif issubclass(obj.__class__, (AutoModel, DelayedModel)):
            key = f"{obj.__class__.__name__}-{obj.pk}"
            if key not in self._save_memo:
                self._save_memo.append(key)
                obj.save()

    def save(self):
        """
        Save this model to the database.

        Returns:
            int: The primary key (pk) of the saved model.
        """
        for attr in self.attributes:
            self._subobj_save(getattr(self, attr))
        self.last_updated = datetime.now()
        record = self.serialize()
        # log(type(record), record)
        self.pk = self.table().save(record)
        return self.pk

    @classmethod
    def get(cls, pk):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """
        # breakpoint()
        result = cls.table().get(pk)
        return cls(**result) if result else None

    @classmethod
    def random(cls):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """
        result = cls.table().random()
        return cls(**result) if result else None

    @classmethod
    def all(cls):
        """
        Get all models of this type.

        Returns:
            list: A list of AutoModel instances.
        """
        return [cls(**o) for o in cls.table().all()]

    @classmethod
    def search(cls, **kwargs):
        """
        Search for models containing the keyword values.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            list: A list of AutoModel instances that match the search criteria.
        """
        return [cls(**attribs) for attribs in cls.table().search(**kwargs)]

    @classmethod
    def find(cls, **kwargs):
        """
        Find the first model containing the keyword values and return it.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            AutoModel or None: The first matching AutoModel instance, or None if not found.
        """
        attribs = cls.table().find(**kwargs)
        return cls(**attribs) if attribs else None

    def _subobj_delete(self, obj):
        if isinstance(obj, list):
            for v in obj:
                self._subobj_delete(v)
        elif isinstance(obj, dict):
            for v in obj.values():
                self._subobj_delete(v)
        elif issubclass(obj.__class__, (AutoModel, DelayedModel)):
            obj.delete()

    def delete(self, sub_objects=True):
        """
        Delete this model from the database.
        """
        if sub_objects:
            for k in self.attributes:
                if attr := getattr(self, k):
                    self._subobj_delete(attr)
        return self.table().delete(pk=self.pk)

    serialized_tracker = []

    @classmethod
    def _serialize(cls, val):
        if isinstance(val, list):
            new_list = []
            for v in val:
                obj = cls._serialize(v)
                new_list.append(obj)
            val = new_list
        elif isinstance(val, dict):
            new_dict = {}
            for k, v in val.items():
                new_dict[k] = cls._serialize(v)
            val = new_dict
        elif isinstance(val, datetime):
            val = {"_datetime": val.isoformat()}
        elif issubclass(val.__class__, (AutoModel, DelayedModel)):
            if val.pk:
                val = {
                    "pk": val.pk,
                    "_automodel": val.model_name(),
                }
            else:
                log(
                    val.__class__.__name__,
                    "The above object was not been saved. You must save subobjects if you want them to persist.",
                )

        return val

    def serialize(self):
        """
        Serialize this model to a dictionary.

        Returns:
            dict: A dictionary representation of the serialized model.
        """
        record = {k: v for k, v in self.__dict__.items() if k in self.attributes}
        result = self._serialize(record)
        return result | {
            "_automodel": self.model_name(),
        }

    @classmethod
    def _deserialize(cls, val):
        if isinstance(val, dict):
            if val.get("_automodel"):
                if val.get("pk"):
                    # log(val.get('pk'))
                    assert len(val.get("pk"))
                    val = DelayedModel(val["_automodel"], val["pk"])
                else:
                    val = None
            elif "_datetime" in val:
                val = datetime.fromisoformat(val["_datetime"])
            else:
                for k, v in val.items():
                    val[k] = cls._deserialize(v)
        elif isinstance(val, list):
            for i, v in enumerate(val):
                val[i] = cls._deserialize(v)
        return val

    @classmethod
    def deserialize(cls, vars):
        """
        Deserialize a dictionary to a model.

        Args:
            vars (dict): The dictionary to deserialize.

        Returns:
            AutoModel: A deserialized AutoModel instance.
        """
        if "_automodel" not in vars:
            raise ValueError("Cannot only deserialize automodel objects")

        return cls(**vars)



"""
    _summary_

_extended_summary_

:return: _description_
:rtype: _type_
"""
import json
import random
import uuid

from redis.commands.json.path import Path
from redis.commands.search.field import NumericField, TagField, TextField
from redis.commands.search.indexDefinition import IndexDefinition, IndexType
from redis.commands.search.query import Query

from autonomous import log
from autonomous.model.autoattribute import AutoAttribute

MAXIMUM_TAG_LENGTH = 1025
replacements = [
    ",",
    ".",
    "<",
    ">",
    "{",
    "}",
    "[",
    "]",
    '"',
    "'",
    ":",
    ";",
    "!",
    "@",
    "#",
    "$",
    "%",
    "^",
    "&",
    "*",
    "(",
    ")",
    "-",
    "+",
    "=",
    "~",
    " ",
]


class Table:
    def __init__(self, name, attributes, db):
        self._db = db
        self.name = name
        # log(attributes)
        self._rules = {}

        for k, v in attributes.items():
            if isinstance(v, AutoAttribute):
                self._rules[k] = v
            elif isinstance(v, str):
                self._rules[k] = AutoAttribute("TAG", default=v)
            elif isinstance(v, (int, float)):
                self._rules[k] = AutoAttribute("NUMERIC", default=v)
            else:
                self._rules[k] = None
        self._index = self._get_index(f"idx:{name}")

    def _get_index(self, name):
        try:
            self._db.ft(name).info()
        except Exception as e:
            # log(e)
            indexed_attrs = []
            for k, options in self._rules.items():
                if options:
                    if options.type == "TEXT":
                        indexed_attrs.append(
                            TextField(f"$.{k}", as_name=k, sortable=True)
                        )
                    elif options.type == "TAG":
                        indexed_attrs.append(
                            TagField(f"$.{k}", as_name=k, sortable=True)
                        )
                    elif options.type == "NUMERIC":
                        indexed_attrs.append(
                            NumericField(f"$.{k}", as_name=k, sortable=True)
                        )
            self._db.ft(name).create_index(
                indexed_attrs,
                definition=IndexDefinition(
                    prefix=[f"{self.name}:"], index_type=IndexType.JSON
                ),
                temporary=60 * 60 * 24 * 7,  # set indexes to expire in 1 week
            )
        return self._db.ft(name)

    def _validate(self, k, v, decode=False, encode=False):
        # log(k, v, self._rules)
        if rule := self._rules.get(k):
            if rule.type in ["TEXT", "TAG"]:
                if decode:
                    for r in replacements:
                        v = v.replace(f"\\{r}", r) if v else ""
                elif encode:
                    if isinstance(v, str):
                        for r in replacements:
                            idx = v.find(r)
                            if idx == 0:
                                v = v.replace(r, f"\\{r}") if v else ""
                            if idx > 0 and v[idx - 1] != "\\":
                                v = v.replace(r, f"\\{r}") if v else ""

                if rule.type == "TAG":
                    try:
                        assert len(v) < MAXIMUM_TAG_LENGTH
                    except AssertionError:
                        raise Exception(
                            f"Invalid attribute value. Must be a less than 1024 characters or use a 'TEXT' option, AutoAttribute('TEXT', default=''): {k}:{v}"
                        )
                    except TypeError as e:
                        v = ""
                        log(f"{e}", f"{k}:{v}")
            elif rule.type == "NUMERIC":
                if v:
                    try:
                        float(v)
                    except TypeError:
                        raise Exception(
                            f"Invalid attribute value. Must be a number: {k}:{v}"
                        )

            if rule.required:
                try:
                    assert v is not None
                except AssertionError:
                    raise Exception(
                        f"Invalid attribute value. Must not be 'None': {k}:{v}"
                    )
        return v

    def save(self, obj):
        # log(obj)
        obj["pk"] = obj.get("pk") or str(uuid.uuid4())
        for k, v in obj.items():
            if rule := self._rules.get(k):
                if rule.type == "TEXT":
                    obj[k] = self._validate(k, v, encode=True)
                else:
                    obj[k] = self._validate(k, v)
        try:
            json_db = self._db.json()
            json_db.set(f"{self.name}:{obj['pk']}", Path.root_path(), obj)
        except TypeError as e:
            log(e)
            raise e
        except Exception as e:
            # log(obj)
            raise e
        obj = {k: self._validate(k, v, decode=True) for k, v in obj.items()}
        return obj["pk"]

    def count(self):
        result = len(self._db.keys(f"{self.name}:*"))
        return result or 0

    def delete(self, pk):
        try:
            return self._db.json().delete(f"{self.name}:{pk}", Path.root_path())
        except Exception as e:
            log(e)
            return None

    def find(self, **search_terms):
        result = self.search(**search_terms)
        return result[0] if result else None

    def search(self, **search_terms):
        # log(search_terms, self.index_name, self.index.info())
        matches = []

        for k, v in search_terms.items():
            try:
                v = self._validate(k, v, encode=True)
            except KeyError:
                raise Exception(f"Can only search indexed fields: {k}:{v}")

            query_str = f"@{k}:{v}"
            ruleset = self._rules[k]
            if ruleset and ruleset.type == "TAG":
                query_str = "@" + str(k) + ":{" + str(v) + "}"
            elif ruleset and ruleset.type == "TEXT":
                query_str = f"@{k}:({v})"

            # log(query_str)

            query = Query(query_str)
            # breakpoint()

            results = self._index.search(query)
            # log(results)
            matches += [json.loads(d.json) for d in results.docs]

        results = {
            obj["pk"]: {k: self._validate(k, v, decode=True) for k, v in obj.items()}
            for obj in matches
        }

        return list(results.values())

    def get(self, pk):
        # log(pk)
        if not pk or pk == "None":
            return None

        if obj := self._db.json().get(f"{self.name}:{pk}", Path.root_path()):
            for k in obj:
                for r in replacements:
                    if isinstance(obj[k], str):
                        obj[k] = obj[k].replace(f"\\{r}", r)
            return obj
        return None

    def all(self):
        keys = self._db.keys(f"{self.name}:*")
        # log(keys)
        objs = self._db.json().mget(keys, Path.root_path()) if keys else []
        for i, obj in enumerate(objs):
            objs[i] = {k: self._validate(k, v, decode=True) for k, v in obj.items()}
        return objs

    def __str__(self):
        return json.dumps(self.all(), indent=4)

    def random(self):
        keys = self._db.keys(f"{self.name}:*")
        key = random.choice(keys).split(":")[-1]
        return self.get(key)

    def clear(self):
        # breakpoint()
        keys = self._db.keys(f"{self.name}:*")
        return self._db.delete(*keys) if keys else None
